#!/usr/bin/env python
# -*- coding: latin-1 -*-

# Imports
import pygame, math, time, random
from pygameutils import make_ball
from numarray import array

# ===========================  BALL  =========================================
class Ball:
	""" Ball that uses angle and motion """
	REPLICATOR = 0
	PREDICTOR = 1
	PLAYERRECT = 2
	LEFTLIMIT = 7
	RIGHTLIMIT = 593
	UPPERLIMIT = 12
	DOWNLIMIT = 288
	SCOREPLAYER1 = 1
	SCOREPLAYER2 = 2
	BOUNCE = 0

	def __init__(self, pos, angle, speed=10):
		""" Inits a ball """
		self.pos = array(pos)
		self.angle = angle
		self.speed = speed
		# We associate the ball with a Rect because we need it for our basic
		# collision detection that are provided by pygame itself
		self.rect = pygame.Rect(pos, (10,10))
		self.lastTime = time.time()

	def updatePosition(self):
		""" Update the position of the ball """
		newTime = time.time()
		dt = newTime - self.lastTime
		x, y = self.pos + self.calculateNewPosition(self.angle, self.speed * dt)
		# Check of ball is outside the game plane
		if x < (self.LEFTLIMIT) and (y > 30 or y < 270):
			self.pos = self.getStartPosition()
			return self.SCOREPLAYER1
		else:
			self.angle = math.pi - self.angle

		if x > (self.RIGHTLIMIT) and (y > 30 or y < 270):
			self.pos = self.getStartPosition()
			return self.SCOREPLAYER2
		else:
			self.angle = math.pi - self.angle

		if y < self.UPPERLIMIT or y > self.DOWNLIMIT:
			self.angle = -self.angle
		newPos = self.calculateNewPosition(self.angle, self.speed * dt)  
		self.pos = self.pos + newPos
		self.lastTime = newTime
		return self.BOUNCE

	def getPosition(self):
		""" Returns the balls position """
		return self.pos

	def getStartPosition(self):
		""" Get the start position of the ball, it will be randomized """
		rand = random.random()
		mult = 7
		if rand < 0.25:
			mult = 1
		elif rand < 0.50:
			mult = 3
		elif rand < 0.75:
			mult = 5

		self.angle = (math.pi/4) * mult
		return array([300, 150])

	def calculateNewPosition(self, angle, speed):
		""" Calculates the new position of the ball, based on the speed and
		    its angle """
		return array([speed * math.cos(angle), speed * math.sin(angle)])

	def doCollisionCheck(self, players):
		""" Does collision detection between the ball and the players' bats.
			If the balls rect and a players rect overlaps, then it will do
			further collision tests to be able to decide if the ball should 
			change direction or not """
		self.rect.center = self.pos
		for player in players.keys():
			# This gives the players rect
			thePlayer = players[player][self.PLAYERRECT]
			if self.rect.colliderect(thePlayer.rect):
				if self.rect.colliderect(thePlayer.leftRect) and \
						math.cos(self.angle) > 0:
							self.angle = math.pi - self.angle
				if self.rect.colliderect(thePlayer.rightRect) and \
						math.cos(self.angle) < 0:
							self.angle = math.pi - self.angle
				if self.rect.colliderect(thePlayer.upRect) and \
						math.sin(self.angle) > 0:
							self.angle = -self.angle
				if self.rect.colliderect(thePlayer.downRect) and \
						math.sin(self.angle) < 0:
							self.angle = -self.angle
		

# ===========================  CLIENTBALL  ===================================
class ClientBall(Ball, pygame.sprite.Sprite):
	""" Class that represensts the ball that a client use """

	def __init__(self, xy, screen, offset):
		""" Inits a ball with its position, xy, its screen and its
		    offset. The offset is used for displaying the second player
			 objects in the lower window """
		pygame.sprite.Sprite.__init__(self)
		Ball.__init__(self, xy, 0, 0)
		self.offset = offset
		self.radius = 10
		self.color = (10, 10, 200)
		self.screen = screen
		self.image, self.rect = make_ball(self.radius, self.color)
		self.rect.center = xy

	def move(self, xy):
		""" Moves the ball by setting the rects coordinates to xy """
		self.rect.center = xy[0], xy[1] + self.offset
