#!/usr/bin/env python
# -*- coding: latin-1 -*-

# Imports
import cPickle, random
from twisted.internet.protocol import DatagramProtocol
from twisted.internet import reactor, protocol
from twisted.spread import pb, jelly
import player
import server
import messages

# ===================    NETWORK          ========================
class Network(DatagramProtocol):
	""" Base class for network handling """
	
	def __init__(self):
		pass
	
	def startProtocol(self):
		""" Called when the network is started. Tries to get the IP at 
		    the localhost """
		print 'Network started...'
		reactor.resolve('localhost').addCallback(self.getIP)

	def getIP(self, ip):
		""" Prints the IP server is using and store this in a local variable """
		self.ip = ip
		print 'Using IP', ip
		return ip

	def startListening(self):
		""" Makes the network handler start listening for traffic at the
		    specified port """
		reactor.listenUDP(self.port, self)

# ===================    SETTINGCOMMUNICATOR ========================
class Communicator(pb.Root):
	""" Communication class """
	def __init__(self, owner):
		print 'Init a Communicator with owner', owner
		self.owner = owner

	def remote_changeServerFramerate(self, frameRate):
		""" Calling the framerate method of a player """
		return self.owner.setFramerate(frameRate)

	def remote_connect(self, address):
		""" Connects to address """
		if not self.owner.players.has_key(address):
			self.owner.addPlayer(address)
		else:
			print 'Player with address', address, 'is already connected'
		return 1

class SettingCommunicator(pb.Root):
	""" Class for changning settings over a network connection """
	
	def __init__(self, owner, port):
		""" Initialize a communicator """
		self.owner = owner
		self.port = port

	def startListening(self):
		""" Set up the communicator in listening mode """
		print 'Calling startListening'
		reactor.listenTCP(self.port, pb.PBServerFactory(Communicator(self.owner)))
		reactor.doSelect(1)
		reactor.runUntilCurrent()

# ===================    CLIENTCOMMUNICATOR ========================
class ClientCommunicator(pb.Referenceable):
	""" Class that make it possible for a client to communicate with
		the server """
		
	def __init__(self, owner, port, serverAddress, serverPort):
		""" Initialize a client communicator """
		self.owner = owner
		self.port = port
		self.serverAddress = serverAddress
		self.serverPort = serverPort
		self.factory = pb.PBClientFactory()

	def startListening(self):
		""" Connect and start listening for network events """
		reactor.connectTCP(self.serverAddress, self.serverPort, self.factory)
		reactor.doSelect(1)
		reactor.runUntilCurrent()

	def connect(self, address):
		""" Connect to the server """
		# Get the servers SettingCommunicator object
		d = self.factory.getRootObject() 
		d.addCallback(self.gotConnected, address).addErrback(self.failure)
	
	def gotConnected(self, remote, address):
		""" Call local methods when we have got connected with the server """
		self.remote = remote
		d = self.remote.callRemote("connect", address)
		d.addCallback(self.owner.setConnected).addErrback(self.failure)


	def changeServerFramerate(self, frameRate):
		""" Change the servers framerate, step 1 """
		# Get the servers SettingCommunicator object
		d = self.factory.getRootObject() 
		d.addCallback(self.changeRate, frameRate).addErrback(self.failure)

	def changeRate(self, remote, frameRate):
		""" Change the servers framerate, step 2 """
		self.remote = remote
		d = self.remote.callRemote("changeServerFramerate", frameRate)
		d.addCallback(self.owner.updateFramerate).addErrback(self.failure)

	def failure(self, err):
		""" Print an error message """
		print 'Couldn\'t make TCP command to server, error:\n', err

# ===================    NETWORKSERVER    ========================
class NetworkServer(Network):
	""" The network instance for the server """
	REPLICATOR = 0
	PREDICTOR = 1
	PLAYERRECT = 2

	def __init__(self, server, port):
		""" Init a network with a server and the port it should listen to """
		Network.__init__(self)
		self.server = server
		self.port = port
		self.ip = None
		self.nbrRec = 0
		self.nbrSent = 0

	def datagramReceived(self, data, (host, port)):
		""" Method that is called when data is incoming from the network """
		# print "received data from %s:%d" % (host, port)
		self.nbrRec += 1
		if (self.nbrRec % 1000) == 0:
			print "nbr of rec UDP", self.nbrRec
		try:
			# Get a message by unpickling the incoming data
			message = cPickle.loads(data)
			
			# Call methods depending what kind of message it is
			if isinstance(message, messages.Message):
				message.addr = (host, port) 
				# Update the predictor for the player
				self.server.players[(host, port)][self.PREDICTOR].update(message)
			else:
				# If it couldn't unpickle it's probably so because it were
				# Missing some UDP-packets.
				print 'Missing packets'
		except cPickle.PickleError:
			print 'NetworkServer :: datagramReceived :: Couldn\'t cPickle'

	def sendMessage(self, message):
		""" Sends a message to all the other players """
		self. nbrSent += 1
		if (self.nbrSent % 1000) == 0:
			print "nbr of sent UDP", self.nbrSent
		try:
			data = cPickle.dumps(message)
			# Send the pickled message to all the players
			for playerAddress in self.server.players.keys():
				self.transport.write(data, playerAddress)
		except cPickle.PickleError:
			print 'NetworkServer :: sendMessage :: Couldn\'t cPickle'


# ===================    NETWORKCLIENT    ========================
class NetworkClient(Network):
	""" The network instance for the clients """
	PREDICTOR = 0
	PLAYERRECT = 1
	SCREEN = 2
	OFFSET = 3
	COLOR = 4

	def __init__(self, client, listeningPort, serverAddress, serverPort):
		""" Init a network with a client and the port it should listen to.
			Also sets the servers IP address and the server port that
			the client should send messages to """
		Network.__init__(self)
		self.client = client
		self.port = listeningPort
		self.serverAddr = (serverAddress, serverPort)
		self.dropPercentage = 0.0

	def addDropPercent(self):
		""" Make the client drop more packets """
		self.dropPercentage += 0.01
		if self.dropPercentage > 1.0:
			self.dropPercentage = 1.0
		return self.dropPercentage

	def subDropPercent(self):
		""" Decrease the packet drop """
		self.dropPercentage -= 0.01
		if self.dropPercentage < 0.0:
			self.dropPercentage = 0.0
		return self.dropPercentage

	def datagramReceived(self, data, (host, port)):
		""" Method that is called when data is incoming from the network """
		if random.random() > self.dropPercentage:
			try:
				# Get a message by unpickling the incoming data
				message = cPickle.loads(data)
				# Call methods depending what kind of message it is
				if isinstance(message, messages.SC_BallPosition):
					# Update the ballpredictor
					self.client.ballPredictor.update(message)
				elif isinstance(message, messages.SC_SetPosition):
					# If the message is from the clients itself, then update its
					# batpredictor, else update the other players batpredictors
					if message.sender == (self.ip, self.port):
						self.client.batPredictor.update(message)
					else:
						# Add player to game if he's not there already, this is a
						# little bit ugly. It checks this everytime for every message.
						# Maybe implementing a message that says "connect" instead or
						# something like that XXX
						if not self.client.players.has_key((host, port)):
							self.client.addPlayer((host, port))
						self.client.players[(host, port)][self.PREDICTOR].update(message)
				elif isinstance(message, messages.SC_Score):
					self.client.updateScore(message)
				else:
					# If it couldn't unpickle it's probably so because it were
					# Missing some UDP-packets.
					print 'Missing packets'
			except cPickle.PickleError:
				print 'NetworkClient :: datagramReceived :: Couldn\'t cPickle'

	def sendMessage(self, message):
		""" Sends a message to the server """
		# Only send UDP-packets to server if the player is connected
		if self.client.connected:
			try:
				data = cPickle.dumps(message)
				# Send the pickled message to the server
				self.transport.write(data, self.serverAddr)
			except cPickle.PickleError:
				print 'NetworkClient :: sendMessage :: Couldn\'t cPickle'
