#!/usr/bin/env python
# -*- coding: latin-1 -*-

# Imports
import math, time
import server
import player
import messages
import ringbuffer
import playerstate
from constants import *
from numarray import array, asarray, matrixmultiply
from pygameutils import calculatePosition, calculateAngle, \
		calculateAcceleration, calculateVelocity, calculateAccelerationVectors

class Predictor:
	""" Base class that makes the predicion """
	NOT_UPDATED = 0
	UPDATED = 1

	POSITION = 0
	LINEAR = 1
	QUADRATIC = 2
	RICHARDSON = 3
	CATMULL = 4

	LEFTLIMIT = 7
	RIGHTLIMIT = 593
	UPPERLIMIT = 12
	DOWNLIMIT = 288
	
	def __init__(self, startPos=(320,240)):
		""" Initialize a predictor """
		self.rb = ringbuffer.Ringbuffer()
		self.rb.append(playerstate.PlayerState(startPos, 0, 0, time.time()))
		self.state = self.NOT_UPDATED
		self.timeNow = time.time()
		self.currentPredictor = self.POSITION
		self.predictors = {}
		self.predictors[self.POSITION] = self.simplePositionPrediction
		self.predictors[self.LINEAR] = self.linearPrediction
		self.predictors[self.QUADRATIC] = self.quadraticPrediction
		self.predictors[self.RICHARDSON] = self.richardsonPrediction
		self.predictors[self.CATMULL] = self.catmullRomPrediction
		self.lastTimeForUpdate = time.time()
		self.nextPredictedUpdateTime = 0
		self.doNotPredict = 0
		self.lastdt = 0



	def getPosition(self):
		""" Returns the position that the predictor think is correct """
		# If there isn't any new packets from network, then predict
		x, y = self.predictors[self.currentPredictor]()
		return array([x, y])

	def simplePositionPrediction(self):
		""" Replicates position """
		return self.rb.getFirst().getPosition()
	
	def linearPrediction(self):
		""" Does simple linear prediction """
		#print 'Linerar Prediction'
		last = self.rb.getFirst()
		dt = time.time() - last.getTime()
		return last.getPosition() + calculatePosition(last.getAngle(), last.getVelocity() * dt)
	
	def quadraticPrediction(self):
		""" Do quadratic prediction """
		# NewPosition = OldPosition + Velocity*Time + Acceleration*(time)2
		last = self.rb.getFirst()
		dt = time.time() - last.getTime()
		velocityVector = calculatePosition(last.getAngle(), last.getVelocity() * dt)
		accelerationVector = (calculateAccelerationVectors(last.getAngle(), \
				last.getAcceleration()) * (dt**2)) / 2
		return last.getPosition() + velocityVector + accelerationVector

	def richardsonPrediction(self):
		""" Do richardson prediction """
		playerStates = self.rb.get()[:3]
		xpos = []
		[xpos.append(x.getXPosition()) for x in playerStates]
		ypos = []
		[ypos.append(y.getYPosition()) for y in playerStates]
		times = []
		[times.append(t.getTime()) for t in playerStates]
		t0t1 = times[0]-times[1]
		if abs(t0t1) < 0.0001:
			t0t1 = 0.0001
			
		t0t2 = times[0]-times[2]
		if abs(t0t2) < 0.0001:
			t0t2 = 0.0001

		tt1 = time.time() - times[1]
		if abs(tt1) < 0.0001:
			tt1 = 0.0001

		# Non-uniform timestep
		c0 = ((4.0 / (3.0*t0t1)) - (1.0 / (3.0*t0t2))) * tt1
		c1 = -((4.0*tt1) / (3.0*t0t1)) + 1.0
		c2 = tt1 / (t0t2*3.0)
		h = array([c0, c1, c2])
		return asarray([self.FIRExtrapolator(h, xpos), self.FIRExtrapolator(h, ypos)])

	def catmullRomPrediction(self):
		""" Do Catmull-Rom prediction """
		playerStates = self.rb.get()[:3]
		xpos = []
		[xpos.append(x.getXPosition()) for x in playerStates]
		ypos = []
		[ypos.append(y.getYPosition()) for y in playerStates]

		# Get the dt for the t - first position
		timeNow = time.time()
		dt = (timeNow - self.rb.getFirst().getTime()) / self.lastdt
		if dt < 0.0001:
			dt = 0.0001

		# Non-uniform timestep
		c0 = 2.5
		c1 = -2.0
		c2 = 0.5
		res = asarray([self.FIRExtrapolator(array([c0, c1, c2]), xpos), \
				self.FIRExtrapolator(array([c0, c1, c2]), ypos)])
		x = res[0] - xpos[0]
		y = res[1] - ypos[0]

		return asarray([xpos[0] + x*dt, ypos[0] + y*dt])

	def FIRExtrapolator(self, coeffs, values):
		""" Do FIR extrapolation """
		return matrixmultiply(asarray(coeffs), asarray(values))

	def changePredictor(self):
		print 'Calling changePredictor in predictor'
		self.currentPredictor = (self.currentPredictor + 1) % len(self.predictors)
		return self.currentPredictor

	def updateValues(self, messageTime, newPos):
		""" Updates speed, angle, position and accerleration when a
			new packet arrives from the replicator to the predictor """
		last = self.rb.getFirst()

		mt = float(messageTime)
		lt = float(last.getTime())
		dt = mt - lt
		self.lastdt = dt
		self.lastTimeForUpdate = mt
		self.nextPredictedUpdateTime = mt + dt

		# Get the new velocity
		newSpeed = calculateVelocity(newPos, last.getPosition(), dt)
		if newSpeed == -1:
			newSpeed = last.getVelocity()
		# Get the new acceleration
		newAcc = calculateAcceleration(newSpeed, self.rb.getFirst().getVelocity(), dt)
		if newAcc == -1:
			newAcc = last.getAcceleration()
		# Get the angle
		newAngle = calculateAngle(newPos, last.getPosition())

		# Append the new variables to the ringbuffer
		if self.doNotPredict:
			# This is applied when a player has scored, need to reset the
			# last X positions in the ringbuffer to init position.
			for x in range(0,4):
				self.rb.append(playerstate.PlayerState((300,150), 0, 0, messageTime, newAngle))
			self.doNotPredict = 0
		else:
			self.rb.append(playerstate.PlayerState(newPos, newSpeed, newAcc, messageTime, newAngle))
		self.state = self.UPDATED
		#self.rb.debugRB()

	def debugPrint(self, mess):
		""" Debug method """
		for x in mess:
			print '\n', mess

# =================  BAT PREDICTOR  ===================
class BatPredictor(Predictor):
	""" Class that predicts where the bat should be """
	def __init__(self, startPos=(0,0)):
		""" Init a batpredicor with position and speed """
		Predictor.__init__(self, startPos)

	def update(self, message):
		""" Updates the predictor with the information that is in
			the message that's given as argument to the method """
		if isinstance(message, messages.CS_Movement):
			self.doMoveUpdate(message)
		elif isinstance(message, messages.SC_SetPosition):
			self.doSetPosition(message)
		else:
			print 'Prediction:: Bat :: Unknown message'

	def doSetPosition(self, message):
		""" This updates a predictor based on the position in the
			message """
		self.updateValues(message.time, message.pos)

	def doMoveUpdate(self, message):
		""" This updates a predictor based on the direction in the
			message """
		dir = message.dir
		# This is just to set it as a semi-dummy position
		newPos = self.rb.getFirst().getPosition()
		if dir == LEFT:
			newPos[0] = newPos[0] - PLAYERSPEED
		elif dir == RIGHT:
			newPos[0] = newPos[0] + PLAYERSPEED
		elif dir == UP:
			newPos[1] = newPos[1] - PLAYERSPEED
		elif dir == DOWN:
			newPos[1] = newPos[1] + PLAYERSPEED
		elif dir == STOP:
			self.speed = 0
			return

		self.updateValues(message.time, newPos)


# ================= BALL PREDICTOR ===================
class BallPredictor(Predictor):
	""" Class that predicts where the ball should be """
	def __init__(self):
		Predictor.__init__(self)

	def update(self, message):
		""" Updates the predictor with the information that is in
			the message that's given as argument to the method """
		self.updateValues(message.time, message.pos)

	def setDefaultStartPosition(self):
		""" Set the default position of the ball """
		self.doNotPredict = 1

